from collections.abc import Generator, Iterable, Sequence

from rich.progress import Progress

from liblaf.grapes import timing
from liblaf.grapes.timing import TimerName
from liblaf.grapes.typed import MISSING, MissingType

from ._progress import progress


def track[T](
    iterable: Iterable[T],
    *,
    description: str = "Progress",
    timers: bool | Sequence[TimerName | str] = ["perf"],
    total: float | None = None,
    callback_end: timing.Callback | MissingType | None = MISSING,
    callback_finally: timing.Callback | MissingType | None = MISSING,
) -> Generator[T]:
    prog: Progress = progress()
    if timers is True:
        timers = ["perf"]
    if total is None:
        total = try_len(iterable)
    if timers:
        if callback_end is MISSING:
            callback_end = timing.log_record(depth=6)
        if callback_finally is MISSING:
            callback_finally = timing.log_summary(depth=5)
        iterable: timing.TimedIterable[T] = timing.timer(
            iterable,
            label=description,
            timers=timers,
            callback_end=callback_end,
            callback_finally=callback_finally,
            total=int(total) if total is not None else None,
        )
        with prog:
            yield from prog.track(iterable, total=total, description=description)
    else:
        with prog:
            yield from prog.track(iterable, total=total, description=description)


def try_len(iterable: Iterable) -> int | None:
    try:
        return len(iterable)  # pyright: ignore[reportArgumentType]
    except TypeError:
        return None



import functools
from collections.abc import Callable

import attrs

from liblaf.grapes import pretty
from liblaf.grapes.typed import MISSING, MissingType

from ._base import TimerRecords
from ._callback import log_record
from ._utils import default_if_missing
from .typed import Callback


# `slots=False` is required to make `functools.update_wrapper(...)` work
# ref: <https://www.attrs.org/en/stable/glossary.html#term-slotted-classes>
@attrs.define(slots=False)
class TimedFunction[**P, T](TimerRecords):
    callback_end: Callback | MissingType | None = attrs.field(
        default=MISSING, converter=default_if_missing(log_record(depth=4)), kw_only=True
    )
    _func: Callable[P, T] = attrs.field(alias="func", on_setattr=attrs.setters.frozen)

    def __attrs_post_init__(self) -> None:
        self.label = self.label or pretty.func(self._func).plain or "Function"
        functools.update_wrapper(self, self._func)

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        self._start()
        result: T = self._func(*args, **kwargs)
        self._end()
        return result



from pathlib import Path
from types import FrameType

from loguru._get_frame import get_frame
from rich.style import Style
from rich.text import Text

from liblaf.grapes.typed import PathLike


def location(
    name: str | None,
    function: str | None,
    line: int | None,
    file: PathLike | None = None,
) -> Text:
    text = Text()
    file: Path | None = Path(file or "<unknown>")
    function = function or "<unknown>"
    line = line or 0
    if file.exists():
        text.append(str(name), style=Style(link=file.as_uri()))
        text.append(":")
        text.append(f"{function}:{line}", style=Style(link=f"{file.as_uri()}#{line}"))
    else:
        text.append(f"{name}:{function}:{line}")
    return text


def caller_location(depth: int = 1) -> Text:
    frame: FrameType | None
    try:
        frame = get_frame(depth)
    except ValueError:
        frame = None
    file: str | None = None
    function: str | None = None
    line: int | None = None
    name: str | None = None
    if frame is not None:
        file = frame.f_code.co_filename
        function = frame.f_code.co_name
        line = frame.f_lineno
        name = frame.f_globals.get("__name__")
    text: Text = location(function=function, line=line, name=name, file=file)
    return text



import contextlib
import types
from collections.abc import Callable, Iterable, Sequence
from typing import Self, overload

import attrs

from liblaf.grapes.typed import MISSING, MissingType

from ._base import TimerRecords
from ._callback import log_record
from ._function import TimedFunction
from ._iterable import TimedIterable
from ._time import TimerName
from .typed import Callback


@attrs.define
class Timer(
    contextlib.AbstractAsyncContextManager,
    contextlib.AbstractContextManager,
    TimerRecords,
):
    async def __aenter__(self) -> Self:
        return self.__enter__()

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: types.TracebackType | None,
        /,
    ) -> None:
        return self.__exit__(exc_type, exc_value, traceback)

    def __call__[**P, T](self, func: Callable[P, T]) -> TimedFunction[P, T]:
        return TimedFunction(
            func,
            label=self.label,
            timers=self.timers,
            callback_start=self.callback_start,
            callback_end=self.callback_end,
            callback_finally=self.callback_finally,
        )

    def __enter__(self) -> Self:
        if self.label is None:
            self.label = "With Block"
        if self.callback_end is MISSING:
            self.callback_end = log_record(depth=4)
        self._start()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: types.TracebackType | None,
        /,
    ) -> None:
        self._end()

    def start(self) -> None:
        self._start()

    def end(self) -> None:
        self._end()


@overload
def timer[T](
    iterable: Iterable[T],
    *,
    label: str | None = None,
    timers: Sequence[TimerName | str] = ["perf"],
    callback_start: Callback | MissingType | None = MISSING,
    callback_end: Callback | MissingType | None = MISSING,
    callback_finally: Callback | MissingType | None = MISSING,
    total: int | None = None,
) -> TimedIterable[T]: ...
@overload
def timer(
    *,
    label: str | None = None,
    timers: Sequence[TimerName | str] = ["perf"],
    callback_start: Callback | MissingType | None = MISSING,
    callback_end: Callback | MissingType | None = MISSING,
    callback_finally: Callback | MissingType | None = MISSING,
) -> Timer: ...
def timer[T](
    iterable: Iterable[T] | None = None,
    *,
    label: str | None = None,
    timers: Sequence[TimerName | str] = ["perf"],
    callback_start: Callback | MissingType | None = MISSING,
    callback_end: Callback | MissingType | None = MISSING,
    callback_finally: Callback | MissingType | None = MISSING,
    total: int | None = None,
) -> TimedIterable[T] | Timer:
    if iterable is not None:
        return TimedIterable(
            iterable=iterable,
            label=label,
            timers=timers,
            callback_start=callback_start,
            callback_end=callback_end,
            callback_finally=callback_finally,
            total=total,
        )
    return Timer(
        label=label,
        timers=timers,
        callback_start=callback_start,
        callback_end=callback_end,
        callback_finally=callback_finally,
    )



import json as _json
from pathlib import Path
from typing import Any, override

from liblaf.grapes import path
from liblaf.grapes.typed import PathLike

from ._abc import AbstractSerializer


class JSONSerializer(AbstractSerializer):
    @override
    def load(self, fpath: PathLike, **kwargs) -> Any:
        fpath: Path = path.as_path(fpath)
        with fpath.open() as fp:
            return _json.load(fp, **kwargs)

    @override
    def loads(self, data: str, **kwargs) -> Any:
        return _json.loads(data, **kwargs)

    @override
    def save(self, fpath: PathLike, data: Any, **kwargs) -> None:
        fpath: Path = path.as_path(fpath)
        with fpath.open("w") as fp:
            _json.dump(data, fp, **kwargs)

    @override
    def saves(self, data: Any, **kwargs) -> str:
        return _json.dumps(data, **kwargs)


json = JSONSerializer()
load_json = json.load
loads_json = json.loads
save_json = json.save
saves_json = json.saves



import functools
from typing import TYPE_CHECKING

from .typed import Callback

if TYPE_CHECKING:
    from ._base import TimerRecords


def log_record(
    index: int = -1,
    label: str | None = None,
    depth: int = 1,
    level: int | str = "DEBUG",
) -> Callback:
    return functools.partial(
        _log_record, index=index, label=label, depth=depth, level=level
    )


def log_summary(
    label: str | None = None,
    depth: int = 1,
    level: int | str = "INFO",
) -> Callback:
    return functools.partial(_log_summary, label=label, depth=depth, level=level)


def _log_record(
    timer: "TimerRecords",
    *,
    index: int = -1,
    label: str | None = None,
    depth: int = 1,
    level: int | str = "DEBUG",
) -> None:
    timer.log_record(index=index, label=label, depth=depth, level=level)


def _log_summary(
    timer: "TimerRecords",
    *,
    label: str | None = None,
    depth: int = 1,
    level: int | str = "INFO",
) -> None:
    timer.log_summary(label=label, depth=depth, level=level)



from pathlib import Path
from typing import Unpack

import loguru
from environs import env
from rich.console import Console

from liblaf.grapes import pretty
from liblaf.grapes.logging.filters import Filter, make_filter
from liblaf.grapes.logging.sink import LoguruRichHandler, TracebackArgs
from liblaf.grapes.typed import PathLike


def rich_handler(
    *,
    console: Console | None = None,
    filter_: Filter | None = None,
    traceback: TracebackArgs | None = None,
    **kwargs: Unpack["loguru.BasicHandlerConfig"],
) -> "loguru.HandlerConfig":
    if console is None:
        console = pretty.get_console("stderr")
    filter_ = make_filter(filter_)
    if traceback is None:
        traceback = TracebackArgs(show_locals=True)
    return {
        "sink": LoguruRichHandler(console=console, traceback=traceback),
        "format": "",
        "filter": filter_,
        **kwargs,
    }


def file_handler(
    *,
    file: PathLike | None = None,
    filter_: Filter | None = None,
    traceback: TracebackArgs | None = None,
    **kwargs: Unpack["loguru.BasicHandlerConfig"],
) -> "loguru.HandlerConfig":
    if file is None:
        file = env.path("LOGGING_FILE", default=Path("run.log"))
    console: Console = pretty.get_console(file)
    filter_ = make_filter(filter_)
    if traceback is None:
        traceback = TracebackArgs(show_locals=True)
    return {
        "sink": LoguruRichHandler(console=console, traceback=traceback),
        "format": "",
        "filter": filter_,
        **kwargs,
    }



from pathlib import Path
from typing import Unpack

import loguru
from environs import env

from liblaf.grapes.logging.filters import Filter, make_filter
from liblaf.grapes.typed import PathLike


def jsonl_handler(
    fpath: PathLike | None = None,
    filter_: Filter | None = None,
    **kwargs: Unpack["loguru.FileHandlerConfig"],
) -> "loguru.HandlerConfig":
    if fpath is None:
        fpath = env.path("LOGGING_JSONL", default=Path("run.log.jsonl"))
    filter_ = make_filter(filter_)
    return {"sink": fpath, "filter": filter_, "serialize": True, "mode": "w", **kwargs}



from collections.abc import Callable
from typing import Any, overload

from liblaf.grapes.typed import MISSING


@overload
def default_if_missing(default: Any = MISSING) -> Callable[[Any], Any]: ...
@overload
def default_if_missing(*, factory: Callable) -> Callable[[Any], Any]: ...
def default_if_missing(
    default: Any = MISSING, factory: Callable | None = None
) -> Callable[[Any], Any]:
    if default is MISSING:

        def converter(val: Any) -> Any:
            if val is MISSING:
                return factory()  # pyright: ignore[reportOptionalCall]
            return val
    else:

        def converter(val: Any) -> Any:
            if val is MISSING:
                return default
            return val

    return converter



import collections
import math
import statistics
import textwrap
from collections.abc import Generator, Mapping, Sequence
from typing import overload

import attrs
from loguru import logger

from liblaf.grapes import human
from liblaf.grapes.typed import MISSING, MissingType

from ._callback import log_summary
from ._time import TimerName, get_time
from ._utils import default_if_missing
from .typed import Callback


@attrs.define
class TimerConfig:
    label: str | None = attrs.field(default=None, kw_only=True)
    timers: Sequence[TimerName | str] = attrs.field(
        factory=lambda: ["perf"], kw_only=True, on_setattr=attrs.setters.frozen
    )

    @property
    def default_timer(self) -> str:
        return self.timers[0]


@attrs.define
class BaseTimer(TimerConfig):
    _time_start: dict[str, float] = attrs.field(
        init=False, factory=dict, on_setattr=attrs.setters.frozen
    )
    _time_end: dict[str, float] = attrs.field(
        init=False, factory=dict, on_setattr=attrs.setters.frozen
    )

    def elapsed(self, timer: str | None = None) -> float:
        timer = timer or self.default_timer
        if self._time_start.get(timer, math.inf) < self._time_end.get(timer, -math.inf):
            return self._time_end[timer] - self._time_start[timer]
        return get_time(timer) - self._time_start[timer]

    def _start(self) -> None:
        for timer in self.timers:
            self._time_start[timer] = get_time(timer)

    def _end(self) -> None:
        for timer in self.timers:
            self._time_end[timer] = get_time(timer)

    @property
    def _current_record(self) -> Mapping[str, float]:
        return {
            timer: self._time_end[timer] - self._time_start[timer]
            for timer in self.timers
        }


@attrs.define
class TimerRecords(BaseTimer):
    callback_start: Callback | MissingType | None = attrs.field(
        default=MISSING, kw_only=True
    )
    callback_end: Callback | MissingType | None = attrs.field(
        default=MISSING, kw_only=True
    )
    callback_finally: Callback | MissingType | None = attrs.field(
        default=MISSING, converter=default_if_missing(factory=log_summary), kw_only=True
    )
    _records: dict[str, list[float]] = attrs.field(
        init=False,
        factory=lambda: collections.defaultdict(list),
        on_setattr=attrs.setters.frozen,
    )

    @overload
    def __getitem__(self, index: int) -> Mapping[str, float]: ...
    @overload
    def __getitem__(self, index: str) -> Sequence[float]: ...
    def __getitem__(self, index: int | str) -> Mapping[str, float] | Sequence[float]:
        if isinstance(index, int):
            return self.row(index)
        return self.column(index)

    def __len__(self) -> int:
        return self.count

    @property
    def columns(self) -> Sequence[str]:
        return self.timers

    @property
    def count(self) -> int:
        return self.n_rows

    @property
    def n_columns(self) -> int:
        return len(self.timers)

    @property
    def n_rows(self) -> int:
        return len(self.column())

    def column(self, timer: str | None = None) -> Sequence[float]:
        timer = timer or self.default_timer
        return self._records[timer]

    def human_record(self, index: int = -1, label: str | None = None) -> str:
        label = label or self.label or "Timer"
        text: str = f"{label} > "
        items: list[str] = []
        for timer, value in self.row(index).items():
            human_duration: str = human.human_duration(value)
            items.append(f"{timer}: {human_duration}")
        text += ", ".join(items)
        return text

    def human_summary(self, label: str | None = None) -> str:
        label = label or self.label or "Timer"
        header: str = f"{label} (total: {self.n_rows})"
        if self.n_rows == 0:
            return header
        body: str = ""
        for timer in self.columns:
            body += f"{timer} > "
            human_mean: str = human.human_duration_series(self.column(timer))
            human_median: str = human.human_duration(self.median(timer))
            body += f"mean: {human_mean}, median: {human_median}\n"
        body = body.strip()
        summary: str = header + "\n" + textwrap.indent(body, "    ")
        return summary

    def iter_columns(self) -> Generator[tuple[str, Sequence[float]]]:
        yield from self._records.items()

    def iter_rows(self) -> Generator[Mapping[str, float]]:
        for index in range(self.n_rows):
            yield self.row(index)

    def log_record(
        self,
        index: int = -1,
        label: str | None = None,
        depth: int = 1,
        level: int | str = "DEBUG",
    ) -> None:
        logger.opt(depth=depth).log(level, self.human_record(index=index, label=label))

    def log_summary(
        self,
        label: str | None = None,
        depth: int = 1,
        level: int | str = "INFO",
    ) -> None:
        logger.opt(depth=depth).log(level, self.human_summary(label=label))

    def row(self, index: int) -> Mapping[str, float]:
        return {timer: values[index] for timer, values in self._records.items()}

    # region statistics

    def max(self, timer: str | None = None) -> float:
        return max(self.column(timer))

    def mean(self, timer: str | None = None) -> float:
        return statistics.mean(self.column(timer))

    def median(self, timer: str | None = None) -> float:
        return statistics.median(self.column(timer))

    def min(self, timer: str | None = None) -> float:
        return min(self.column(timer))

    def std(self, timer: str | None = None) -> float:
        return statistics.stdev(self.column(timer))

    # endregion statistics

    def _append(
        self, seconds: Mapping[str, float] = {}, nanoseconds: Mapping[str, float] = {}
    ) -> None:
        for key, value in seconds.items():
            self._records[key].append(value)
        for key, value in nanoseconds.items():
            self._records[key].append(value * 1e-9)

    def _end(self) -> None:
        super()._end()
        self._append(seconds=self._current_record)
        if callable(self.callback_end):
            self.callback_end(self)

    def _start(self) -> None:
        super()._start()
        if callable(self.callback_start):
            self.callback_start(self)



from ._base import BaseTimer, TimerConfig, TimerRecords
from ._callback import log_record, log_summary
from ._function import TimedFunction
from ._iterable import TimedIterable
from ._time import TimerName, get_time
from ._timer import Timer, timer
from .typed import Callback

__all__ = [
    "BaseTimer",
    "Callback",
    "TimedFunction",
    "TimedIterable",
    "Timer",
    "TimerConfig",
    "TimerName",
    "TimerRecords",
    "get_time",
    "log_record",
    "log_summary",
    "timer",
]
